package grm;

import interfaces.Asct;
import interfaces.AsctHelper;
import interfaces.ExecutionManager;
import interfaces.GrmPOA;
import interfaces.Lrm;
import interfaces.LrmHelper;

import subtypes.AsctRequestId;
import types.CommonExecutionSpecs;
import types.DistinctExecutionSpecs;
import types.DynamicInfo;
import types.StaticInfo;

import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.StringTokenizer;

import org.omg.CORBA.ORB;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.ORBPackage.InvalidName;
import org.omg.CosTrading.IllegalConstraint;
import org.omg.CosTrading.IllegalServiceType;
import org.omg.CosTrading.Lookup;
import org.omg.CosTrading.LookupHelper;
import org.omg.CosTrading.OfferIteratorHolder;
import org.omg.CosTrading.OfferSeqHolder;
import org.omg.CosTrading.Policy;
import org.omg.CosTrading.PolicyNameSeqHolder;
import org.omg.CosTrading.Property;
import org.omg.CosTrading.Register;
import org.omg.CosTrading.UnknownServiceType;
import org.omg.CosTrading.LookupPackage.HowManyProps;
import org.omg.CosTrading.LookupPackage.SpecifiedProps;
import org.omg.CosTrading.RegisterPackage.NoMatchingOffers;
import org.omg.CosTrading.RegisterPackage.OfferInfo;
import org.omg.CosTradingRepos.ServiceTypeRepository;
import org.omg.CosTradingRepos.ServiceTypeRepositoryHelper;
import org.omg.CosTradingRepos.ServiceTypeRepositoryPackage.PropStruct;
import org.omg.CosTradingRepos.ServiceTypeRepositoryPackage.PropertyMode;
import org.omg.CosTradingRepos.ServiceTypeRepositoryPackage.TypeStruct;

  ///
  /// Class GrmImpl - Servant implementation of GRM.idl
  ///
  /// The GRM acts as a resource availability descriptions repository and
  /// applications execution request scheduler
  ///
  ///@author Andrei Goldchleger
  ///
  public class GrmImpl extends GrmPOA implements Runnable{

    private ExecutionManager execMonitor = null;

    private static final int NUM_STATIC_FEATURES = 8;/*< Node static features
                                                        (e.g. OS, CPU class)
                                                      */
    private static final int NUM_DYNAMIC_FEATURES = 7;/*< Node dynamic features
                                                         (e.g. % free CPU)
                                                       */

    private static final int NUM_FEATURES = /*< all features */
                              NUM_STATIC_FEATURES + NUM_DYNAMIC_FEATURES + 2;


    private int execNumber = 0;

    private ORB orb_;         ///< reference to the ORB
    private Lookup lookup_;   ///< reference to the ORB lookup interface
    private HashMap regLrm_;  ///< hashMap that keeps id of registered LRMs
    ServiceTypeRepository serviceTypeRepository_; ///< reference to the
                                                  ///  ServiceTypeRepository interface

    ///------------------------------------------------------------------------------
    ///
    /// Creates a new GrmImpl object
    ///
    /// @param orb - a reference to an ORB, used to access CORBA services
    ///
    public GrmImpl(ORB orb){

      regLrm_ = new HashMap();
      orb_ = orb;
      try{
        lookup_ = LookupHelper.narrow(orb_.resolve_initial_references("TradingService"));
      }
      catch(InvalidName in){
        in.printStackTrace();
      }
      serviceTypeRepository_ =  ServiceTypeRepositoryHelper.narrow(lookup_.type_repos());

      if(typeExists("NodeInfo"))
        removeAllOffers("NodeInfo");
      else
        createNodeInfoServiceType();

      (new Thread(this)).start();
    }

    ///------------------------------------------------------------------------------
    ///
    /// Removes all offers of a given service type from the trader
    ///
    /// @param serviceType - the string name of the service type
    ///
    private void removeAllOffers(String serviceType){
      //Removes all existing offers from trader(cleanup)
      try{
        Register res = lookup_.register_if();
        res.withdraw_using_constraint(serviceType, "TRUE");
      }
      catch (NoMatchingOffers nmo){//TODO: I believe that this exception
                                   //is being launched when there are no
                                    //offers in  the trader
        System.out.println("GrmImpl::removeAllOffers->Empty trader");
      }
      catch (UnknownServiceType ust){
        ust.printStackTrace();
        System.exit(1);
      }
      catch (IllegalServiceType ist) {
        ist.printStackTrace();
        System.exit(1);
      }
      catch (IllegalConstraint ic){
        ic.printStackTrace();
        System.exit(1);
      }
    }

    ///------------------------------------------------------------------------------
    ///
    /// Checks if a given service type exists in the trader
    ///
    /// @param serviceType - the string name of the service type
    ///
    /// @returns true if the service type exists, false otherwise
    ///
    private boolean typeExists(String serviceType){

      TypeStruct type_desc;
      try {
        type_desc = serviceTypeRepository_.describe_type(serviceType);
        System.out.println("Service type \"" +
                           serviceType +
                           "\" exists");
        return true;
      }

      catch (UnknownServiceType ust) {
        System.out.println("Service type \"" +
                           serviceType +
                           "\" does not exist");
        return false;
      }
      catch (IllegalServiceType ist) {
        ist.printStackTrace();
        System.exit(1);
      }
      return false;
    }

    ///------------------------------------------------------------------------------
    ///
    /// Creates the NodeInfo Service Type, which will hold information about the grid
    /// machines
    ///
    private void createNodeInfoServiceType(){

      PropStruct [] ps = new PropStruct[NUM_FEATURES];

      TypeCode tcs = orb_.get_primitive_tc(TCKind.from_int(TCKind._tk_string));
      TypeCode tci = orb_.get_primitive_tc(TCKind.from_int(TCKind._tk_long));
      TypeCode tcb = orb_.get_primitive_tc(TCKind.from_int(TCKind._tk_boolean));
      PropertyMode mandatory = PropertyMode.from_int(PropertyMode._PROP_MANDATORY);
      PropertyMode normal = PropertyMode.from_int(PropertyMode._PROP_NORMAL);
      ps[0] = new PropStruct("lrmIOR",tcs,mandatory);
      ps[1] = new PropStruct("hostName",tcs,mandatory);
      ps[2] = new PropStruct("osName",tcs,mandatory);
      ps[3] = new PropStruct("osVersion",tcs,mandatory);
      ps[4] = new PropStruct("processorName",tcs,mandatory);
      ps[5] = new PropStruct("processorMhz",tci,mandatory);
      ps[6] = new PropStruct("totalRAM",tci,mandatory);
      ps[7] = new PropStruct("totalSwap",tci,mandatory);
      ps[8] = new PropStruct("fsTotal",tci,normal);
      ps[9] = new PropStruct("filesTotal",tci,normal);
      ps[10] = new PropStruct("freeRAM",tci,normal);
      ps[11] = new PropStruct("freeSwap",tci,normal);
      ps[12] = new PropStruct("fsFree",tci,normal);
      ps[13] = new PropStruct("filesFree",tci,normal);
      ps[14] = new PropStruct("cpuUsage",tcs,normal);
      ps[15] = new PropStruct("recentlyPicked",tcb,mandatory);
      ps[16] = new PropStruct("lastUpdated",tci,mandatory);

      try{
        serviceTypeRepository_.add_type("NodeInfo",//FIXME: Put string in a constant
                       "IDL:Lrm:1.0",  //FIXME: Is this the correct value ???
                       ps,
                       new String []{});
      }
      catch(Exception e){
         e.printStackTrace();
         System.exit(1);
       }
    }

    ///------------------------------------------------------------------------------
    ///
    /// Registers a LRM on this GRM
    ///
    /// @param lrmId - LRM's IOR
    /// @param staticInfo - Object representing a node static info
    ///
    public void registerLrm(String lrmId, StaticInfo staticInfo){


      System.out.println("GrmImpl::registerLrm");
      System.out.println("Lrm: " + lrmId);
      dumpStaticInfo(staticInfo);

      //verifies if the LRM is already registered. Trader does not avoid
      //duplicate entries on its database
      synchronized(regLrm_){
        if(regLrm_.get(lrmId) != null){
          System.out.println("GrmImpl::register_Lrm->Attempt to double register a LRM");
          return;
        }
      }

      //Registers LRM in trader
      //FIXME: We are not addressing unusual situations, such as two LRMs with the
      //FIXME: same ID. Since the id is the LRM's IOR, this should be impossible
      //FIXME: to happen
      Property [] props = new Property[NUM_STATIC_FEATURES + 2];
      props[0] = new Property("lrmIOR",orb_.create_any());
      props[0].value.insert_string(lrmId);
      props[1] = new Property("hostName",orb_.create_any());
      props[1].value.insert_string(staticInfo.hostName);
      props[2] = new Property("osName",orb_.create_any());
      props[2].value.insert_string(staticInfo.osName);
      props[3] = new Property("osVersion",orb_.create_any());
      props[3].value.insert_string(staticInfo.osVersion);
      props[4] = new Property("processorName",orb_.create_any());
      props[4].value.insert_string(staticInfo.processorName);
      props[5] = new Property("processorMhz",orb_.create_any());
      props[5].value.insert_long(staticInfo.processorMhz);
      props[6] = new Property("totalRAM",orb_.create_any());
      props[6].value.insert_long(staticInfo.totalRAM);
      props[7] = new Property("totalSwap",orb_.create_any());
      props[7].value.insert_long(staticInfo.totalSwap);
      props[8] = new Property("recentlyPicked",orb_.create_any());
      props[8].value.insert_boolean(false);
      props[9] = new Property("lastUpdated",orb_.create_any());
      props[9].value.insert_long((int) (new Date()).getTime() / 1000);

      try{
        String offerId = lookup_.register_if().export(orb_.string_to_object(lrmId),
                                                      "NodeInfo",
                                                      props);
        System.out.println("Offerid: "+ offerId);
        synchronized(regLrm_){
          regLrm_.put(lrmId, offerId);
        }
      }
      catch(Exception e){
        e.printStackTrace();
      }
    }//Method

    ///------------------------------------------------------------------------------
    ///
    /// Receive an information update from a given LRM
    ///
    /// @param lrmId - LRM's IOR
    /// @param dynamicInfo - Object representing a node dynamic info
    ///
    public void sendUpdate(String lrmId, DynamicInfo dynamicInfo){

      //We cannot update an unregistered LRM
      synchronized(regLrm_){
        if(regLrm_.get(lrmId) == null){
          System.out.println("GrmImpl::send_update->Unregistered  Lrm ");
          return;
        }
      }

      Property [] props = new Property[NUM_DYNAMIC_FEATURES + 2];
      props[0] = new Property("freeRAM",orb_.create_any());
      props[0].value.insert_long(dynamicInfo.freeRAM);
      props[1] = new Property("freeSwap",orb_.create_any());
      props[1].value.insert_long(dynamicInfo.freeSwap);
      props[2] = new Property("fsFree",orb_.create_any());
      props[2].value.insert_long(dynamicInfo.fsFree);
      props[3] = new Property("filesFree",orb_.create_any());
      props[3].value.insert_long(dynamicInfo.filesFree);
      props[4] = new Property("cpuUsage",orb_.create_any());
      props[4].value.insert_string(dynamicInfo.cpuUsage);
      props[5] = new Property("fsTotal",orb_.create_any());
      props[5].value.insert_long(dynamicInfo.fsTotal);
      props[6] = new Property("filesTotal",orb_.create_any());
      props[6].value.insert_long(dynamicInfo.filesTotal);
      props[7] = new Property("recentlyPicked",orb_.create_any());
      props[7].value.insert_boolean(false);
      props[8] = new Property("lastUpdated",orb_.create_any());
      props[8].value.insert_long((int) (new Date()).getTime() / 1000);

      try{
        synchronized(regLrm_){
          lookup_.register_if().modify((String) regLrm_.get(lrmId),
                                      new String[] {},
                                       props);
        }
      }
      catch(Exception e){
        e.printStackTrace();
      }
    }

    ///------------------------------------------------------------------------------
    ///
    /// Receives a remote execution request.
    ///
    /// Receives a remote execution request from an ASCT and finds suitable
    /// machines to service that request, if available. If not, notifies
    /// the requesting ASCT.
    ///
    /// @param commonSpecs - common specs to all application copies
    /// @param distinctSpecs - specific specs regarding a given application copy. In
    /// the case of sequential applications, there is only one spec
    /// @param forceDifferentMachines - if true, the GRM will only service the request
    /// if EACH of the application copies can be executed on a different machine
    ///
    public void remoteExecutionRequest(CommonExecutionSpecs commonSpecs,
                                        DistinctExecutionSpecs[] distinctSpecs,
                                        boolean forceDifferentMachines){

      //Merge constraints and visited(i.e. for a given query, do not visit a node twice)
      //FIXME: Note that we expect constraints to be adequately formatted TCL queries
      String constraints = generateConstraints(commonSpecs.applicationConstraints,
                                               commonSpecs.deniedExecution);
      String preferences = generatePreferences(commonSpecs.applicationPreferences);

      // getting the list of implemented binary files of this application
      // vidal 02/2005
      String binaryConstraint = "";
      String result = "";
      String os_name = "";
      String processor_name = "";
      
      //System.out.println("commonSpecs.appIds.length = "+commonSpecs.appIds.length);

      //generate a constraint for specific platform(s)
      for(int i = 0; i < commonSpecs.appIds.length; i++) {
      	//System.out.println("appId["+i+"] = "+commonSpecs.appIds[i]);

      	int pos = commonSpecs.appIds[i].indexOf("_");
      	os_name = commonSpecs.appIds[i].substring(0,pos);
      	processor_name = commonSpecs.appIds[i].substring(pos+1);
      	result  = "(osName == '" + os_name + "' and " + "processorName == '" + processor_name + "')";
      	binaryConstraint = binaryConstraint + result ;
      	if (i < commonSpecs.appIds.length -1 )
	      binaryConstraint = binaryConstraint +  " or ";
      }
      if(constraints.length()>0) {
      	constraints = constraints + " and ";
      }
      constraints = constraints + binaryConstraint;    
      
      
      //generate a OfferSeqHolder (osh) from a query using the constraint above
      System.out.println("GrmImpl::multiRemoteExecutionRequest");
      System.out.println("Constraints: " + constraints);
      System.out.println("Preferences: " + preferences);
      OfferSeqHolder osh = new OfferSeqHolder();
      try{
        SpecifiedProps sp = new SpecifiedProps();
        sp.__default(HowManyProps.some);
        sp.prop_names(new String [] {"lrmIOR","osName","processorName"}); // vidal 02/2005
        lookup_.query("NodeInfo",
                      constraints,
                      preferences,
                      new Policy[] {},//default policies
                      sp,
                      distinctSpecs.length,
                      osh,
                      new OfferIteratorHolder(),
                      new PolicyNameSeqHolder());
      }
      catch(Exception e){//FIXME: Catch each exception separately, and some of them should sent
        e.printStackTrace();//messages back to lrms
      }

      if(((osh.value.length < distinctSpecs.length) && forceDifferentMachines)
         || (osh.value.length == 0)){//No host suitable to execute application
        System.out.println("Needed: "
                           + distinctSpecs.length
                           + " hosts. Got: " + osh.value.length + " hosts.");
        Asct asct = AsctHelper.narrow(orb_.string_to_object(commonSpecs.requestingAsctIor));
        for(int i = 0; i < distinctSpecs.length; i++)
          asct.refusedExecutionRequest(distinctSpecs[i].asctRequestId);
        return;
      }//if
          
      ArrayList offersIors = new ArrayList();

      //vidal 02/2005
      String originalBinarylPathId = commonSpecs.applicationId;
      //System.out.println("osh.value.length> "+ osh.value.length);
      //System.out.println("1) applicationId> "+ originalBinarylPathId );
 
      // associate compatibles <binaryId,lrmIOR>
      String [][] lrmBinaryPair = new String[osh.value.length][];
      for(int i = 0; i < osh.value.length; i++) {
      	//System.out.println("osh.value[" + i + "].properties.length> "
      	//		+osh.value[i].properties.length);
      	
      	String os = osh.value[i].properties[1].value.extract_string();
      	String hard = osh.value[i].properties[2].value.extract_string();
      	String platform = os + "_" + hard;
	  
      	//System.out.println(osh.value[i].properties[0].value.extract_string());
      	lrmBinaryPair[i] = new String[2];
      	lrmBinaryPair[i][0] = osh.value[i].properties[0].value.extract_string();

      	/* gets the right binaryId (the meaning of Id TextFields on the AsctGui 
      	 * changes whether we choice executeApplication or 
      	 * executeBinary on the AsctGui) */
      	if (commonSpecs.isExecuteApp) {
      		// executeApplication
      		lrmBinaryPair[i][1] =  originalBinarylPathId+ "/"+ platform+"_binaryFile";
      	}
  		else {
  			// executeBinary 
  			lrmBinaryPair[i][1] = commonSpecs.applicationId;
  		}
      	offersIors.add(lrmBinaryPair[i]);
      	//System.out.println("2) applicationId> "+ lrmBinaryPair[i][1] );
      }//for

//      StringBuffer newAsctAppId = new StringBuffer();
//      try{              
//          newAsctAppId.append(java.net.InetAddress.getLocalHost().getHostAddress().replaceAll("[.]",""));
//          newAsctAppId.append(Calendar.getInstance().getTimeInMillis());
//          newAsctAppId.append(":");
//          newAsctAppId.append(++execNumber);
//          newAsctAppId.append(":");
//          newAsctAppId.append(distinctSpecs[0].asctRequestId.appMainRequestId);
//      }
//      catch (UnknownHostException e) {
//          System.err.println("Could not get host Info. Aborting request...");            
//          e.printStackTrace();
//      }

      StringBuffer newAsctAppId = new StringBuffer();
      if (distinctSpecs[0].asctRequestId.appMainRequestId.indexOf(":") < 0 ) {
          try{              
              newAsctAppId.append(java.net.InetAddress.getLocalHost().getHostAddress().replaceAll("[.]",""));
              newAsctAppId.append(Calendar.getInstance().getTimeInMillis());
              newAsctAppId.append(":");
              newAsctAppId.append(++execNumber);
              newAsctAppId.append(":");              
          }
          catch (UnknownHostException e) {
              System.err.println("Could not get host Info. Aborting request...");            
              e.printStackTrace();
          }
      }
      newAsctAppId.append(distinctSpecs[0].asctRequestId.appMainRequestId);

      // Send a message to the ExecutionManager containing the request properties
      for (int i=0; i < distinctSpecs.length; i++)
            distinctSpecs[i].asctRequestId.appMainRequestId = newAsctAppId.toString();  
      if (execMonitor != null)
          execMonitor.reportExecutionScheduled(commonSpecs, distinctSpecs, forceDifferentMachines);

      //since compatibles <binaryId,lrmIOR> are associated
      //we can call lrm.remoteExecutionRequest(commonSpecs, distinctSpecs[i])
      //with the right applicationId 
      int i = 0;
      while(i < distinctSpecs.length){
        try{
          //distinctSpecs[i].asctRequestId.appMainRequestId = newAsctAppId.toString();  
          String[] pair = (String [])offersIors.get(i % offersIors.size());
          String lrmId = pair[0];
          String binaryId = pair[1];
          Lrm lrm = LrmHelper.narrow( orb_.string_to_object(lrmId) );
          commonSpecs.applicationId = binaryId;
          lrm.remoteExecutionRequest(commonSpecs, distinctSpecs[i]);
          touchLrm(lrmId);
          i++;
        }
        catch(org.omg.CORBA.TRANSIENT transientException){
          System.err.println("Unable to reach LRM. Removing from database");
          removeLrm((String) offersIors.get(i % offersIors.size()));
          synchronized(regLrm_){
             regLrm_.remove((String)offersIors.get(i % offersIors.size()));
          }
          offersIors.remove(i % offersIors.size());
        }      
      }//while
      
      //FIXME: Here 2k catches Sys and User exceptions and withdraw offer. Consider
      //doing that also
    }//method


    ///------------------------------------------------------------------------------
    ///
    /// Periodically Checks if the registered LRMs are online.
    ///
    /// For each registered LRM, verifies if the time since the last information update
    /// does not exceed a defined threshold. If it exceeds, ping the LRM in order to determine
    /// if it is online
    ///
    public void run(){

      while(true){
        Set keys = null;
        String [] lrmIors = null;
        synchronized(regLrm_){
          keys = regLrm_.keySet();
          Iterator it = keys.iterator();
          lrmIors = new String[keys.size()];
          for(int i = 0; i < keys.size(); i++)
            lrmIors[i] = (String) it.next();
        }

        for(int i = 0; i < lrmIors.length; i++){

          OfferSeqHolder osh = new OfferSeqHolder();
          try{
            SpecifiedProps sp = new SpecifiedProps();
            sp.__default(HowManyProps.some);
            sp.prop_names(new String [] {"lastUpdated"});
            lookup_.query("NodeInfo",
                      "lrmIOR == '" + lrmIors[i] + "'",//constraints
                      "",//Preferences
                      new Policy[] {},
                      sp,
                      1,//We are interested in one suitable offer
                      osh,
                      new OfferIteratorHolder(),
                      new PolicyNameSeqHolder());
          }
          catch(Exception e){//FIXME: Catch each exception separately, and some of them should sent
            e.printStackTrace();//messages back to lrms
          }
          int lastUpdated = osh.value[0].properties[0].value.extract_long();
          int currentTime =  (int) (new Date()).getTime() / 1000;
          if((currentTime - lastUpdated) > 200){//FIXME: Define interval somewhere
            Lrm lrm = LrmHelper.narrow(orb_.string_to_object(lrmIors[i]));
            try{
              lrm.ping();
            }
            catch(org.omg.CORBA.TRANSIENT transientException){
              System.err.println("GrmImpl::Run-->Unable to reach LRM. Removing from database. lrmIor: "+
              lrmIors[i]);
              removeLrm(lrmIors[i]);
              synchronized(regLrm_){
                regLrm_.remove(lrmIors[i]);
              }
            }
          }//if
        }//for
        try{
          Thread.sleep(1000 * 30);//FIXME: Define interval somewhere
        }catch(java.lang.InterruptedException ie){
          ie.printStackTrace();
        }
      }//while(true)
    }

    ///------------------------------------------------------------------------------
    
    public void registerExecutionManager(ExecutionManager execMonitor){
        this.execMonitor = execMonitor;
    }
    
    ///------------------------------------------------------------------------------
    ///
    /// Modify the offer associated to a given LRM in order to indicate that the LRM
    /// was recently chosen to service an execution request
    ///
    /// @param lrmIor - the IOR of the given LRM
    ///
    private void touchLrm(String lrmIor){
             Property [] props = new Property[1];
       props[0] = new Property("recentlyPicked", orb_.create_any());
       props[0].value.insert_boolean(true);
       try{
         Register reg = lookup_.register_if();
         synchronized(regLrm_){
           reg.modify((String)regLrm_.get(lrmIor),
                      new String[] {},
                      props);
         }
       }
       catch(Exception e){//FIXME: Maybe it is better to catch individual exceptions
         e.printStackTrace();
       }
    }

    ///------------------------------------------------------------------------------
    ///
    /// Removes an offer from the trader
    ///
    /// @param lrmIor - the IOR of a LRM associated to the offer
    ///
    private void removeLrm(String lrmIor){
       try{
         Register reg = lookup_.register_if();
         synchronized(regLrm_){
           reg.withdraw((String)regLrm_.get(lrmIor));
         }
       }
       catch(Exception e){//FIXME: Maybe it is better to catch individual exceptions
         e.printStackTrace();
       }
    }

    ///------------------------------------------------------------------------------
    ///
    /// Modifies a TCL query in order to prefer machines which were not
    /// recently selected to service an execution request
    ///
    /// @param preferences - original TCL preferences string
    ///
    /// @returns the modified TCL query
    ///
    private String generatePreferences(String preferences){
      if (preferences.length() > 0)
        preferences ="with " + preferences + " and recentlyPicked != TRUE";
      else
        preferences = "with recentlyPicked != TRUE";
      return preferences;
    }

    ///------------------------------------------------------------------------------
    ///
    /// Generates a TCL constraint that describes all machines that already
    /// denied an execution request
    ///
    /// @param constraints - the original constraints
    /// @param denyingHosts - space separated string containig ids of machines
    /// that denied to host a given execution request
    ///
    /// @returns the properly formatted TCL string representing the constraint
    ///
    private String generateConstraints(String constraints, String denyingHosts){
      String result = constraints.trim();
      if((denyingHosts.length() > 0) && (result.length() > 0))
         result = result + " and ";
      StringTokenizer st  = new StringTokenizer(denyingHosts);
      if(st.hasMoreTokens())
        result = "lrmIOR != '" + st.nextToken() + "'";
      while(st.hasMoreTokens())
        result.concat(" and lrmIOR != '" + st.nextToken()+ "'");
      return result.trim();
    }// generateConstraints method

    ///------------------------------------------------------------------------
    private void dumpLrm(String offerId){

      try{
        System.out.println("GrmImpl::dumpLrm");
        Register res = lookup_.register_if();
        OfferInfo oi = res.describe(offerId);
        for(int i = 0; i < oi.properties.length; i++)
          System.out.println(oi.properties[i].name+":  "+oi.properties[i].value);
      }
      catch(Exception e){
        e.printStackTrace();
      }
    }

    ///-----------------------------------------------------------------------------
    private void dumpStaticInfo(StaticInfo staticInfo){

       System.out.println("HostName: "      +staticInfo.hostName);
       System.out.println("OsName: "        +staticInfo.osName);
       System.out.println("OsVersion: "     +staticInfo.osVersion);
       System.out.println("ProcessorName: " +staticInfo.processorName);
       System.out.println("ProcessorMhz: "  +staticInfo.processorMhz);
       System.out.println("TotalRAM: "      +staticInfo.totalRAM);
       System.out.println("TotalSwap: "     +staticInfo.totalSwap);
    }

    ///-----------------------------------------------------------------------------
    private void dumpDynamicInfo(DynamicInfo dynamicInfo){

      System.out.println("FreeRAM: "       + dynamicInfo.freeRAM);
      System.out.println("FreeSwap: "      + dynamicInfo.freeSwap);
      System.out.println("FsFree: "        + dynamicInfo.fsFree);
      System.out.println("FilesFree: "     + dynamicInfo.filesFree);
      System.out.println("CpuUsage: "      + dynamicInfo.cpuUsage);
      System.out.println("FsTotal: "       + dynamicInfo.fsTotal);
      System.out.println("FilesTotal: "    + dynamicInfo.filesTotal);
    }

  }//class





