#include "CheckpointingLib.hpp"
#include "CheckpointTimer.hpp"
#include "CkpCompilerStack.hpp"
#include "CkpRestoredData.hpp"
#include "HeapController.hpp"
#include "CkpStore.hpp"

#include "utils/c++/Config.hpp"
#include "utils/c++/StringUtils.hpp"

#include <iostream>
#include <cassert>

extern "C" {
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
}

using namespace std;

CheckpointTimer *ckpTimer = NULL;

void initCheckpointing(int ckpInterval) {
  if (ckpTimer == NULL)
    ckpTimer = new CheckpointTimer(ckpInterval);
}

int testCheckpointTimer() {

  if (ckpTimer == NULL)
    return 0;

  if (ckpTimer->testFinishedInterval())
    return 1;
  
  return 0;
}

int saveCheckpoint(const char *filename, void *data, long nbytes) {

  // Creates the File
  int outFile = creat(filename,0666);
  if (outFile == -1) 
    return -1;

  ckp_datasize_t dataSize = (ckp_datasize_t)nbytes;
  // Writes the data to the file
  write(outFile, &dataSize, sizeof(ckp_datasize_t));
  int writenBytes = write(outFile, data, nbytes);
  close(outFile);
  
  return writenBytes;
}

int recoverCheckpoint(const char *filename, void **data, long* bufferSize) { 
	
  int inFile = open(filename, O_RDONLY);
  if (inFile == -1) 
    return -1;

  //  ckp_datasize_t dataSize;
  //  read(inFile, &dataSize, sizeof(ckp_datasize_t));
  struct stat *fileStat = new struct stat;
  fstat(inFile, fileStat);
  *bufferSize = (long)fileStat->st_size;
  delete fileStat;

  *data = malloc(*bufferSize);
  ckp_datasize_t readBytes = read(inFile, *data, *bufferSize);

  assert ((long)readBytes == *bufferSize);

  return (int)readBytes;
}

//------------------------------------------------------------------------------
// Used by ckpCompiler
//------------------------------------------------------------------------------

int ckp_restoring = 0;
CkpStore *ckpStore = NULL;     // initialized at ckp_initialize_stack
CkpCompilerStack *ckpStack = NULL; // initialized at ckp_initialize_stack
CkpRestoredData  *ckpRestoredData = NULL;
HeapController heapController;

void createCkpStore(Config & ckpConfig);
//-------

// FIXME: Should not need the 'initial_capacity' parameter
void ckp_initialize_stack(int initial_capacity) {

  if (ckpStore) delete ckpStore;
  ckpStore = NULL;

  // Try to locate a "ckp.conf" config file in the current directory
  try { 
    Config ckpConfig("ckp.conf"); 
    createCkpStore(ckpConfig);
  } 
  catch(...) { // A file was not created
    // Data is saved to the local directory with no delay
    ckpStore = new CkpFileStore(string(""), 0);
  }     
  
  // Initializes the checkpointing stack
  ckpStack = new CkpCompilerStack(initial_capacity, &heapController, ckpStore); 
}

void createCkpStore(Config & ckpConfig) {

  try {

    string ckpStoreType = ckpConfig.getConf("ckpStore");
    int ckpInterval;

    if (ckpStoreType.compare("noCheckpointing") == 0) {
      std::cerr << "createCkpStore --> NoCheckpointing" << std::endl;
      ckpInterval = 60*60*24*365; // a year =)    
      ckpStore = new CkpNoStore();    
    }
    else if (ckpStoreType.compare("ckpNoStore") == 0) {
      ckpInterval = StringUtils::string2int(ckpConfig.getConf("interval"));
      ckpStore = new CkpNoStore();    
      std::cerr << "createCkpStore --> CkpNoStore " << std::endl;
    }
    else if (ckpStoreType.compare("ckpLocalStore") == 0) {
      ckpInterval = StringUtils::string2int(ckpConfig.getConf("interval"));
      string ckpPath = ckpConfig.getConf("location");
      ckpStore = new CkpFileStore(ckpPath, 0);
      std::cerr << "createCkpStore --> CkpLocalStore " << ckpPath << std::endl;
    }
    else if (ckpStoreType.compare("ckpReposStore") == 0) {
      ckpInterval = StringUtils::string2int(ckpConfig.getConf("interval"));
      ostringstream repPathStream;
      repPathStream << ckpConfig.getConf("execId")  << " " 
		    << ckpConfig.getConf("address") << " " 
		    << ckpConfig.getConf("port");
      ckpStore = new CkpRepositoryStore(repPathStream.str(), 0);
      std::cerr << "createCkpStore --> CkpReposStore " << repPathStream.str() << std::endl;
    }

    if (ckpTimer) delete ckpTimer;
    ckpTimer = new CheckpointTimer(ckpInterval);

  }
  catch(...) { // NoSuchConfigException
    std::cerr << "ERROR: createCkpStore --> " 
	      << "Invalid parameters setting checkpointing storage type." << std::endl;
  }

  
}

// Deprecated
void set_ckp_store(int ckpSaveFlag, int delay) {

  if (ckpStore) delete ckpStore;

  if (ckpSaveFlag == -1) /** Data is saved to the local directory with no delay */
    ckpStore = new CkpFileStore("", delay);
  else if (ckpSaveFlag == -10)
    ckpStore = new CkpFileStore("/tmp/", delay);
  else if (ckpSaveFlag == -20)
    ckpStore = new CkpNoStore();
  else if (ckpSaveFlag == -30)
    std::cerr << "ERROR: set_ckp_store --> Functionality not implemented." << std::endl;
  
  ckpStack->setCkpStore(ckpStore); 

}

int ckp_push_data(void *data, long nbytes, unsigned long type, void (*func)(void *)) {
    return ckpStack->ckpPushData(data, nbytes, type, func);
}

int ckp_pop_data(void) {
    return ckpStack->ckpPopData(1);
}

int ckp_npop_data(int nVar) {
    return ckpStack->ckpPopData(nVar);
}

int ckp_save_stack_data(int ckpNumber) {
  int returnValue = ckpStack->ckpSaveStackData(ckpNumber);  
  sleep(1);
  return returnValue;
}

//-------

int ckp_restore_data(int ckpNumber) {

    if (ckpRestoredData != NULL) delete ckpRestoredData;
    ckpRestoredData = new CkpRestoredData;

    int bytesRead = ckpRestoredData->ckpRestoreCkpData(ckpStore, ckpNumber);
    ckpStack->setLastCkpNumber(ckpNumber);
    return bytesRead;
}

int ckp_get_data(void *data, long nbytes, unsigned long type, void (*func)(void *)) {
    return ckpRestoredData->ckpGetData(data, nbytes, type, func);
}

//------------------------------------------------------------------------------
// Memory Management
//------------------------------------------------------------------------------

void *ckp_malloc(size_t size) {
  void *address = malloc(size);
  heapController.addAllocatedMemory(address, size);  
  return address;
}

void *ckp_calloc(size_t nmenb, size_t size) {
  void *address = calloc(nmenb, size);
  heapController.addAllocatedMemory(address, size);  
  return address;
}

void ckp_free(void *ptr) {
  free(ptr);
  heapController.removeAllocatedMemory(ptr);  
}

void *ckp_realloc(void *ptr, size_t size) {
  heapController.removeAllocatedMemory(ptr);      
  void *address = realloc(ptr, size);
  heapController.addAllocatedMemory(address, size);  
  return address;
}

