extern "C" {
#include <pthread.h>
#include <lualib.h>
#include <lauxlib.h>
#include <lua.h>
}

#include "LrmSkeleton.hpp"
#include "LrmImpl.hpp"
#include "LinuxSpecifics.hpp"

#include "utils/c++/Config.hpp"
#include "utils/c++/OrbUtils.hpp"
#include "utils/c++/LuaUtils.hpp"
#include "utils/c++/StringUtils.hpp"

#include "ExecutionSpecs.hpp"

using std::cerr;
using std::endl;



  LrmSkeleton * LrmSkeleton::singleInstance_ = NULL;

  //-----------------------------------------------------------------------
  LrmSkeleton & LrmSkeleton::init(LrmImpl * lrmImpl_, const Config & config){

    if(LrmSkeleton::singleInstance_ == NULL){
      LrmSkeleton::singleInstance_ = new LrmSkeleton(lrmImpl_, config);
      pthread_t thread;
        pthread_create(&thread, NULL,(void * (*)(void *))
                        serverSideSetup ,(void *) NULL );
      pthread_detach(thread);

    }
    return *LrmSkeleton::singleInstance_;
  }

  //-----------------------------------------------------------------------
  void * LrmSkeleton::serverSideSetup(void * ptr){
  	lua_State *state = LrmSkeleton::singleInstance_->serverSideState;
    LuaUtils::getField(state,"oil.run");
    if(lua_pcall(state, 0, 0, 0) != 0){
      cerr << "[ERROR] LrmSkeleton::serverSideSetup->Lua error: "
           << lua_tostring(state, -1) << endl;
      lua_pop(state, 1);
    }
    return NULL;
  }

  //---------------------------------------------------------------------
  LrmSkeleton::LrmSkeleton(LrmImpl * lrmImpl,
                             const Config & config):
                                                   lrmImpl_(lrmImpl){
    serverSideState = lua_open();
    OrbUtils::initLuaState(serverSideState);
    OrbUtils::loadOrb(serverSideState, config.getConf("orbPath"));
    OrbUtils::loadIdl(serverSideState, config.getConf("resourceManagementIdlPath"));

    lua_register(serverSideState, "setSampleIntervalWrapper", setSampleIntervalWrapper);
    lua_register(serverSideState, "setKeepAliveIntervalWrapper", setKeepAliveIntervalWrapper);
    lua_register(serverSideState, "remoteExecutionRequestWrapper", remoteExecutionRequestWrapper);
    lua_register(serverSideState, "requestOutputFilesWrapper", requestOutputFilesWrapper);
    lua_register(serverSideState, "getLastCkpNumberWrapper", getLastCkpNumberWrapper);
    lua_register(serverSideState, "restartExecutionWrapper", restartExecutionWrapper);    
    lua_register(serverSideState, "killWrapper", killWrapper);

    //servant implementation
    lua_dostring(serverSideState, " lrmImpl = {"
                        " setSampleInterval = function (self, seconds)"
                        "   setSampleIntervalWrapper(seconds)"
                        " end,"
                        " setKeepAliveInterval = function (self, seconds)"
                        "   setKeepAliveIntervalWrapper(seconds)"
                        " end,"
                        " remoteExecutionRequest = function (self, commonSpecs, distinctSpecs)"
                        "   return remoteExecutionRequestWrapper(commonSpecs, distinctSpecs)"
                        " end,"
                        " requestOutputFiles = function(self, appId)"
                        "   return requestOutputFilesWrapper(appId)"
                        " end,"
                        " kill  = function(self, appId)"
                        "   killWrapper(appId)"
                        " end,"
                        " getLastCkpNumber = function(self, appId)"
                        "   return getLastCkpNumberWrapper(appId)"
                        " end,"
                        " restartExecution = function(self, appId, applicationArgs)"
                        "   return restartExecutionWrapper(appId, applicationArgs)"
                        " end,"
                        " ping = function(self)"
                        "   print('Getting pinged...')"
                        " end,"
                        " }"
                );

    lua_dostring(serverSideState,
                 "lrmServant = oil.createservant(lrmImpl, 'IDL:interfaces/Lrm:1.0')");
    lrmIor_ = OrbUtils::getIor(serverSideState, "lrmServant");
    cout << "LrmSkeleton::serverSideSetup::LRM IOR: " << lrmIor_ << endl;
  }

  //---------------------------------------------------------------------
  int LrmSkeleton::setSampleIntervalWrapper(struct lua_State * state){
    int n = lua_gettop(state);
    int sampleIntervalTmp = int(lua_tonumber(state, n));
    LrmSkeleton::singleInstance_->lrmImpl_->setSampleInterval(sampleIntervalTmp);
    return 0;
  }

  //---------------------------------------------------------------------
  int LrmSkeleton::setKeepAliveIntervalWrapper(struct lua_State * state){

    int n = lua_gettop(state);
    int keepAliveIntervalTmp = int(lua_tonumber(state, n));
    LrmSkeleton::singleInstance_->lrmImpl_->setKeepAliveInterval(keepAliveIntervalTmp);
    return 0;
  }

  //---------------------------------------------------------------------
  int LrmSkeleton::remoteExecutionRequestWrapper(struct lua_State * state){
    ExecutionSpecs *execSpecs = new ExecutionSpecs;
    //Here we have a table DistinctExecutionSpecs at the top of the stack
    lua_pushstring(state, "asctRequestId");
    lua_gettable(state, -2);
    execSpecs->appMainRequestId(LuaUtils::getStringFromTable(state,"appMainRequestId"));
    execSpecs->appNodeRequestId(LuaUtils::getStringFromTable(state,"appNodeRequestId"));
    lua_pop(state, 1);//removes asctRequestId table
    execSpecs->appArgs(LuaUtils::getStringFromTable(state,"applicationArgs"));
    lua_pushstring(state, "outputFiles");
    lua_gettable(state, -2);
    execSpecs->outputFiles(LuaUtils::extractStringSequence(state, -1));
    lua_pop(state, 2); //removes outputFiles table & DistinctExecutionSpecs table
    execSpecs->requestingAsctIor(LuaUtils::getStringFromTable(state,"requestingAsctIor"));
    execSpecs->grmIor(LuaUtils::getStringFromTable(state,"grmIor"));
    execSpecs->deniedExecution(LuaUtils::getStringFromTable(state,"deniedExecution"));
    execSpecs->appId(LuaUtils::getStringFromTable(state,"applicationId"));
    execSpecs->appConstraints(LuaUtils::getStringFromTable(state,"applicationConstraints"));
    execSpecs->appPreferences(LuaUtils::getStringFromTable(state,"applicationPreferences"));
    lua_pop(state, 1); //removes CommonExecutionSpecs table

    long test = 0;
    test = LrmSkeleton::singleInstance_->lrmImpl_->remoteExecutionRequest(execSpecs);
    lua_pushnumber(state, test);

    return 1;
  }

  //---------------------------------------------------------------------
  int LrmSkeleton::requestOutputFilesWrapper(struct lua_State * state){
    cout << "LrmSkeleton::requestOutputFilesWrapper" << endl;
    string appId(lua_tostring(state,-1));
    lua_pop(state, 1);//Pops appId
    const AppInfoHolder appInfo = LrmSkeleton::singleInstance_->lrmImpl_->appInfo(appId);
    appInfo.dump();
    lua_newtable(state);
    int lastSetFile = 1;
    for(unsigned int i = 0; i < appInfo.outputFiles().size(); i++){
      cout << "i: " << i << endl;
      lua_newtable(state);
      LuaUtils::setFieldOnTable(state, "filename", StringUtils::stripPath(appInfo.outputFiles()[i]).c_str(), -1);
      lua_pushstring(state, "file");
      try{
        string appPath = appId + "/" +  StringUtils::stripPath(appInfo.outputFiles()[i]);
        LuaUtils::openFileForRead(state, appPath);
        LuaUtils::readFullFile(state);        
        lua_settable(state, -3);
        lua_rawseti(state, -2, lastSetFile++);
        LuaUtils::closeDefaultFile(state);
      }
      //FIXME: This is being added for the case when a ASCT request files that do not
      //necessarily exist in a LRM. This happens, for example, In BSP executions where
      //just one node outputs a given file. If possible, it will be better to fix the
      //design thus avoiding these situations. Alternatively, we should at least warn
      //the requesting ASCT via a CORBA exception, ant let the calling asct decide if
      //it is an error condition.
      catch(const std::exception & e){
        cerr << "LrmSkeleton::outputFilesWrapper -->>\"Silent Fail\" "
             << " -- Someone asked for an inexistent file: "
             << appInfo.outputFiles()[i]
             << endl;
        lua_pop(state, 2);//Popping Table & 'file' string
      }

    }
    LrmSkeleton::singleInstance_->lrmImpl_->removeAppInfo(appId);
    return 1;
  }

  //---------------------------------------------------------------------
  int LrmSkeleton::killWrapper(struct lua_State * state){
    cout << "LrmSkeleton::kill" << endl;
    string appId(lua_tostring(state,-1));
    lua_pop(state, 1);//Pops appId
    LrmSkeleton::singleInstance_->lrmImpl_->kill(appId);
    return 0;
  }

  //---------------------------------------------------------------------
  int LrmSkeleton::getLastCkpNumberWrapper(struct lua_State * state){
    //cout << "LrmSkeleton::getLastCkpNumber" << endl;
    string appId(lua_tostring(state,-1));
    lua_pop(state, 1);//Pops appId
    int ckpNumber = LrmSkeleton::singleInstance_->lrmImpl_->getLastCkpNumber(appId);
    cout << "LrmSkeleton::getLastCkpNumber-->ckpNumber: " << ckpNumber << endl;
    lua_pushnumber(state, ckpNumber);    
    return 1;
  }

  //---------------------------------------------------------------------
  int LrmSkeleton::restartExecutionWrapper(struct lua_State * state){
    cout << "LrmSkeleton::restartExecution" << endl;
    string *appArgs = new string(lua_tostring(state,-1));
    lua_pop(state, 1);// pops appArgs
    string *appId = new string(lua_tostring(state,-1));
    lua_pop(state, 1);// pops appId    
    int restartFlag = LrmSkeleton::singleInstance_->lrmImpl_->restartExecution(appId, appArgs);
    lua_pushnumber(state, restartFlag);
    return 1;
  }

